#pylint: disable=C0103

from builtins import range
def allchoice(L):
    pool = [[]]
    for i in range(0, len(L)):
        np = []
        for j in range(0, len(pool)):
            for k in range(0, len(L[i])):
                np.append(pool[j] + [L[i][k]])
        pool += np
    return pool


# print allchoice([[1,2],[3,4],[5,6,7]])



from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals
from __future__ import absolute_import
from builtins import super
from builtins import range
from builtins import open
from builtins import str
from future import standard_library
standard_library.install_aliases()

import requests
import sys
import bs4
import datetime
import pickle
import hashlib
import math
import os.path
import io
import praw
import configparser
from .ocr import ocrcomp
from enum import Enum
from collections import deque


class Origins(Enum):
    """ Enum for holding the origins of memes.
    """

    NA = 0
    """ Representing an unknown origin.
    """

    QUICKMEME = 1
    """ Representing quickmeme.com.
    """

    MEMEGENERATOR = 2
    """ Representing memegenerator.net.
    """

    REDDITMEMES = 3
    """ Represeting Reddit /r/meme subreddit.
    """

    @classmethod
    def string_to_enum(self, s):
        """ Conver string to a Origins Enum object
        Args:
            s (str): The string representing the name of the origin
        """
        if s.lower() == "quickmeme":
            return self.QUICKMEME
        elif s.lower() == "memegenerator":
            return self.MEMEGENERATOR
        elif s.lower() == "redditmemes":
            return self.REDDITMEMES
        else:
            return self.NA


class Meme(object):
    """ A class for representing memes

    This class provides a high-level abstraction for memes.

    **Attributes:**
        * _pic_url (str): A string representing the url of the picture
        * _caption (str): A string representing the caption of the meme
        * _time (datetime object): The time of creation of the meme
        * _origin (Orgins Enum): The origins enum object representing
          the origin
        * _tags (list): A list of string representing the categories
          of the meme
    """

    def __init__(self, pic_url, time,
                 title=None,
                 caption=None,
                 raw_pic_url=None,
                 origin=Origins.NA,
                 tags=[],
                 score=-1):
        """ __init__ method for Meme class


        :param str pic_url: URL of the meme picture
        :param datetime time: Time of creation
        :param str caption: The caption of the meme
        :param str raw_pic_url: The url of the picture without caption
        :param origin: The origin of the meme (website)
        :param list tags: A list of strings representing tags
        :param int score: A score representing the popularity of the meme
        :type origin: Origins (Enum Type)
        """
        self._pic_url = pic_url
        self._time = time
        self._title = title
        self._caption = caption
        self._raw_pic_url = raw_pic_url
        self._origin = Origins.NA
        self._tags = tags

    def get_pic_url(self):
        """ Get url to the picture

        :return: The url to the meme picture. Notice that this picture
            contains the captions.
        :rtype: str
        """
        return self._pic_url

    def get_caption(self):
        """ Get caption of the meme

        :return: The captions of the meme.
        :rtype: str
        """
        return self._caption

    def get_title(self):
        """ Get the title of the meme

        :return: The title of the meme
        :rtype: str
        :raises ValueError: if the meme does not have a title
        """
        if self._title is None or len(self._title) == 0:
            raise ValueError("Meme does not have a title.")
        return self._title

    def get_time(self):
        """ Return the meme's creation time

        :return: The creation time of the meme
        :rtype: datetime object
        """
        return self._time

    def get_raw_pic_url(self):
        """ Return the url of the meme's picture without caption

        :return: The url pointing to the meme's background picture
        :rtype: str
        :raises ValueError: if the meme does not have a empty
            background picture
        """
        if self._raw_pic_url is None or len(self._raw_pic_url) == 0:
            raise ValueError("Meme does not have a background picture url.")
        return self._raw_pic_url

    def get_origin(self):
        """ Return the origin of the meme
        """
        return self._origin

    def get_tags(self):
        """ Representing a list of tags for the meme
        """
        return self._tags

    def ocr_caption(self, method="Tesseract", **kwargs):
        """ Use ocr to update self caption

        **OCR Methods Available**

        * `Tesseract <https://github.com/tesseract-ocr/tesseract>`:
          Open-source OCR Engine
        * FontMatching: Using Impact Font and template matching to conduct OCR

        When using Tesseract, users need to provide two keyword arguments:

            * thres (bool): a boolean indicating whether we need to threshold
              the image
            * cfg (str): a string representing the configuration to use
              for Tesseract

        """

        def checkKwargs():
            if kwargs is None:
                raise ValueError(
                    "Please provide Tesseract method kwargs.")
            else:
                try:
                    thres = kwargs["thres"]
                    cfg = kwargs["cfg"]

                    if type(thres) is not bool:
                        raise ValueError(
                            "Threshold value must be a boolean.")

                    if type(cfg) is not str:
                        raise ValueError(
                            "Configuration name must be a string.")

                except KeyError:
                    raise KeyError("Legal entries: thres and cfg.")

        # Create a file-like object using Requests and BytesIO
        r = requests.get(self._pic_url, stream=True)
        path = io.BytesIO(r.content)

        if self._caption is None or len(self._caption) == 0:
            # run ocr routine
            if method == "Tesseract":
                checkKwargs()
                print("Now performing OCR with"
                      " Tesseract and {}".format(str(kwargs)))
                result = ocrcomp.ocrTesseract(
                    path, thres=kwargs["thres"], cfg=kwargs["cfg"])

                self._caption = result
            elif method == "FontMatching":
                result = ocrcomp.ocr(path)
                self._caption = result
            elif method == "Auto":
                checkKwargs()
                A = ocrcomp.ocrcomp(path, ocrcomp.ocr,
                                    lambda x: ocrcomp.ocrTesseract(
                                        x, thres=True, cfg=kwargs["cfg"]),
                                    lambda x: ocrcomp.ocrTesseract(
                                        x, thres=True, cfg="Default"),
                                    lambda x: ocrcomp.ocrTesseract(
                                        x, thres=False, cfg=kwargs["cfg"]),
                                    lambda x: ocrcomp.ocrTesseract(
                                        x, thres=False, cfg="Default"))
                self._caption = A[-1][-1]
            else:
                print(method)
                raise ValueError("Not a supported mathod. Methods available: "
                                 "Tesseract, FontMatching, Auto")
        else:
            print("Caption already exists.")

    def __hash__(self):
        """ Two memes are the same if they have the same urls and the same capture time
        """
        return hash((self._pic_url, self._time))

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return (self._pic_url == other._pic_url and
                    self._time == other._time)
        return NotImplemented

    def __ne__(self, other):
        if isinstance(other, self.__class__):
            return not self == other
        return NotImplemented

    def __repr__(self):
        return "Meme URL:{:s} Time:{!s} Caption:{!s} Origin:{!s} Tags:{!s}"\
            .format(self._pic_url,
                    self._time,
                    self._caption,
                    self._origin,
                    self._tags)


class MemeSite(object):
    """ A base class for any sites with respect to memes

    This class should be subclassed. The MemeSite is designed to keep
    all Memes in a cache file, so that even if the Python process is
    terminated, the next time we run the save process, we don't need
    to re-download all the memes from the Internet. The _meme_pool
    and _meme_deque store memes, but the users should not view the memes
    in them as constant, as operations on the object will change the memes
    inside the pool and deque.

    Attributes:
        _url (str): URL for the website hosting memes
        _max_tries (int): Max tries for http requests
        _meme_pool (set): A set containing stored memes
        _meme_deque (deque): A deque containing stored memes
        _last_update (datetime object): The time of last download of memes
        _cache_size (int): Number of memes stored on disk
        _maxcache_day (int): Max day of keeping the cache on disk
    """

    def __init__(self, url, cache_size=500, maxcache_day=1):
        self._url = url
        self._max_tries = 10
        self._meme_pool = set()
        self._meme_deque = deque()
        self._last_update = datetime.datetime.now()
        self._cache_size = cache_size
        self._maxcache_day = maxcache_day

        try:
            self._main_page = requests.get(url)
        except Exception as err:
            sys.stderr.write("ERROR: {} \n".format(str(err)))

    def get_captions(self, num_memes):
        """ Return a list of captions
        """
        a = self.get_memes(num_memes)
        return [x.get_caption() for x in a]

    def get_memes(self, num_memes):
        """ Return a list of memes
        """
        raise NotImplementedError("Implement in subclasses.")

    def clean_meme_pool(self):
        """ Empty the meme pool
        """
        self._meme_pool = set()

    def clean_meme_deque(self):
        """ Empty the meme deque
        """
        self._meme_deque.clear()

    def get_url(self):
        """ Return the base url
        """
        return self._url

    def get_meme_pool(self):
        """ Return a set of memes
        """
        return self._meme_pool

    def get_meme_num(self):
        """ Return the number of memes we have
        """
        return len(self._meme_deque)

    def get_unique_meme_num(self):
        """ Return the number of unique memes we have
        """
        return len(self._meme_pool)

    def _pop_memes(self, n):
        """ Pop number of memes out of the pool
        Return a list of memes
        """
        r = []
        for i in range(n):
            r.append(self._meme_deque.pop())
        return r

    def _read_cache(self):
        """ Read the saved cache file (no side effects)
        Read a tuple containing the data
        """

        if os.path.isfile(self._filepath()):
            # Read the file in using Pickle
            file_obj = open(self._filepath(), 'rb')
            data = pickle.load(file_obj)
            # print(data)
            # self._read_data_tuple(data)
            file_obj.close()
            return data
        else:
            file_obj.close()
            raise OSError("No cache exists.")

    def _update_with_cache(self):
        """ Update self states with cache
        """
        data = self._read_cache()
        self._read_data_tuple(data)

    def _save_cache(self):
        """ Save the class to cache files

        We put all the variables into a tuple, and then save that tuple
        using pickle.
        """
        # Store all the variables into a tuple
        # Save the data to a file with a unique name

        file_obj = open(self._filepath(), 'wb')
        pickle.dump(self._write_data_tuple(), file_obj)
        file_obj.close()

    def _cache_expired(self):
        """ Check whether cache has expired. Also return false when cache doesn't exist
        """
        try:
            delta_time = datetime.datetime.now() \
                - self._read_update_time_from_cache()
            result = delta_time > datetime.timedelta(days=self._maxcache_day)

            if result:
                print("Cache for {} has expired.".format(self._url))
            else:
                print("Cache for {} is not expired.".format(self._url))

            return result

        except OSError:
            print("Cache for {} is not expired.".format(self._url))
            return False

    def _no_cache(self):
        """ Check whether cache exists
        """
        result = os.path.isfile(self._filepath())

        if result:
            print("Cache for {} exists.".format(self._url))
        else:
            print("Cache for {}  does not exist.".format(self._url))

        return not result

    def _build_cache(self):
        """ Build cache
        """
        print("Building cache for {}.".format(self._url))
        self._populate(self._cache_size)
        self._save_cache()

    def _populate(self):
        """ Populate the meme pool and deque
        """
        raise NotImplementedError("Implement in subclasses")

    def _read_data_tuple(self, t_data):
        """ Read in a data tuple and replace all the instance variables
        """
        raise NotImplementedError("Implement in subclasses.")

    def _read_update_time_from_cache(self):
        """ Read cache update time from cache file
        """
        raise NotImplementedError("Implement in subclasses.")

    def _write_data_tuple(self):
        """ Write all internal states to a data tuple
        """
        raise NotImplementedError("Implement in subclasses.")

    def _filename(self):
        """ Use SHA1 hashing algorithm to calculate a unique file name
        """
        raise NotImplementedError("Implement in subclasses.")

    def _filepath(self):
        """ Generate the path to the cache file
        """
        cdir = os.path.dirname(os.path.realpath(__file__))
        tgt_path = os.path.join(cdir, self._filename())
        return tgt_path

    def __repr__(self):
        return "Memesite URL:{:s} Pool:{!s} Update Time:{!s}"\
            .format(self._url,
                    self._meme_pool,
                    self._last_update)


class QuickMeme(MemeSite):
    """ The MemeSite subclass that deals with the quickmeme site.

    quickmeme.com uses an infinite scrolling homepage.
    Fortunately, we can also access the later pages by
    just going to the url:

    www.quickmeme.com/page/i/, where i is the page number
    each page contains 10 user posts; each post consists
    of an image and an alternative text
    """

    def __init__(self, cache_size=500, maxcache_day=1):
        super(QuickMeme, self).__init__(
            "http://www.quickmeme.com/", cache_size, maxcache_day)
        self._posts_per_page = 10
        self._origin = Origins.QUICKMEME

        if self._no_cache() or self._cache_expired():
            self._build_cache()

    def get_memes(self, num_memes):
        """
        Get a number of memes from Quickmeme.com
        """
        # Check the time difference and whether the
        # cache has been created
        if self._no_cache() or self._cache_expired():
            self._build_cache()
        else:
            # Read in saved memes
            self._update_with_cache()

        # Check whether we have enough memes
        if self._cache_size >= num_memes:
            return self._pop_memes(num_memes)
        else:
            # Find the page number of the last page
            pnum = math.ceil(num_memes / self._posts_per_page)

            curl = self._url + "page/{:d}/".format(pnum)
            cpage = requests.get(curl)
            csoup = bs4.BeautifulSoup(cpage.text, 'html.parser')
            # Extract posts from current page
            meme_posts = csoup.find_all(
                class_="post-image", limit=num_memes % self._posts_per_page)

            # Extract captions and picture urls from posts
            texts = [str(x['alt']).rpartition("  ") for x in meme_posts]
            urls = [str(x['src']) for x in meme_posts]

            # Get the additional memes
            # Start from the meme after the last one in cache
            for i in range(self._cache_size % self._posts_per_page,
                           len(meme_posts)):
                time = datetime.datetime.now()
                meme = Meme(urls[i], time, caption=texts[i][0],
                            origin=self._origin, tags=[texts[i][-1]])
                self._meme_pool.add(meme)
                self._meme_deque.appendleft(meme)

            result = self._pop_memes(self.num_memes)
            return result

    def _populate(self, num):
        """ Populate the meme pool and deques
        """
        # Each page on quickmeme contains 10 meme posts
        # So the number of pages to crawl is:
        # num_memes / posts_per_page + mod ( num_memes, posts_per_page)
        max_page = math.ceil(num / self._posts_per_page)
        for i in range(1, max_page + 1):
            if i != max_page:
                self._memes_on_page(i, self._posts_per_page)
            else:
                self._memes_on_page(i, num % self._posts_per_page)

        # Current date and time
        self._last_update = datetime.datetime.now()

    def _memes_on_page(self, page_num, n):
        """Get n memes from page_num page

        Remarks:
        For the meme deque, we put memes in from the left side.
        If we want to retrieve memes from the most popular to the least,
        we pop from the right side (FIFO). If we want to retrieve memes
        the other way around, we pop from the left side (FILO).

        We also use set so that we can keep a unique collection of memes.

        Args:
        page_num -- the number of page we would like to retrieve
        n -- the number of memes we would like to retrieve from that page
        """
        if n > self._posts_per_page:
            return None

        curl = self._url + "page/{:d}/".format(page_num)
        cpage = requests.get(curl)
        csoup = bs4.BeautifulSoup(cpage.text, 'html.parser')
        # Extract posts from current page
        meme_posts = csoup.find_all(class_="post-image", limit=n)

        # Extract captions and picture urls from posts
        texts = [str(x['alt']).rpartition("  ") for x in meme_posts]
        urls = [str(x['src']) for x in meme_posts]

        # Populate the _meme_pool
        for i in range(len(meme_posts)):
            time = datetime.datetime.now()
            meme = Meme(urls[i], time, texts[i][0],
                        self._origin, [texts[i][-1]])
            self._meme_pool.add(meme)
            self._meme_deque.appendleft(meme)

    def _filename(self):
        """ Use SHA1 hashing algorithm to calculate a unique file name
        """
        hashID = hashlib.sha1()
        hashID.update(repr(self._url).encode('utf-8'))

        # Create a file name with hexdecimal representation of the SHA1 hash
        file_name = "cache_{:s}.memecache".format(hashID.hexdigest())
        return file_name

    def _read_data_tuple(self, t_data):
        """ Read in a data tuple and replace all the instance variables
        """
        self._url = t_data[0]
        self._max_tries = t_data[1]
        self._meme_pool = t_data[2]
        self._meme_deque = t_data[3]
        self._last_update = t_data[4]
        self._cache_size = t_data[5]
        self._maxcache_day = t_data[6]

    def _read_update_time_from_cache(self):
        """ Read cache update time from cache file
        """
        dtuple = self._read_cache()

        return dtuple[4]

    def _write_data_tuple(self):
        """ Write all internal states to a data tuple
        """
        data = (self._url, self._max_tries, self._meme_pool,
                self._meme_deque, self._last_update, self._cache_size,
                self._maxcache_day)
        return data


class MemeGenerator(MemeSite):
    """ This class represents the memegenerator.net website
    """

    def __init__(self, cache_size=500, maxcache_day=1,
                 popular_type="Daily", timeout=20):
        """ The __init__ method for MemeGenerator class

        Args:
            cache_size (int): Number of memes stored as cache
            maxcache_day (int): Number of days until the cache expires
        """
        super(MemeGenerator, self).__init__(
            "http://www.memegenerator.net", cache_size, maxcache_day)
        self._origin = Origins.MEMEGENERATOR
        self._api = "http://version1.api.memegenerator.net/"
        self._method_entry = "Instances_Select_ByPopular"

        if popular_type == "Daily":
            self._popular_days = 1
        elif popular_type == "Weekly":
            self._popular_days = 7
        elif popular_type == "Monthly":
            self._popular_days = 30
        else:
            raise ValueError(
                "Wrong popular type. Supported: Daily, Weekly, Monthly")

        self._timeout = timeout
        self._posts_per_page = 15
        if self._no_cache() or self._cache_expired():
            self._build_cache()

    def get_memes(self, num_memes):
        """ Get a number of memes from memegenerator.net
        """

        if self._no_cache() or self._cache_expired():
            self._build_cache()
        else:
            self._update_with_cache()

        if self._cache_size >= num_memes:
            return self._pop_memes(num_memes)
        else:
            # Caculate the additional pages we need to scrap to get the
            # remaining memes.
            result = self._pop_memes(self._cache_size)
            pnum = math.ceil(num_memes / self._posts_per_page)
            print("Last page: ", pnum)
            first_pnum = math.ceil(self._cache_size / self._posts_per_page)
            additional_memes = []

            # Get all the additional memes
            # Notice that we are double counting some of our original cached
            # memes and counting more memes than we needed on the last page
            for i in range(first_pnum, pnum + 1):
                additional_memes = additional_memes + self._get_memes_helper(i)

            # Get ride of the additional memes and resolve the double counting
            # issue
            pre_index = self._cache_size % self._posts_per_page
            waste = self._posts_per_page * pnum - num_memes

            additional_memes = additional_memes[
                pre_index:len(additional_memes) - waste]

            result = result + additional_memes
            return result

    def _populate(self, num):
        """ Populate the meme pool and deque
        """
        max_page = math.ceil(num / self._posts_per_page)
        for i in range(1, max_page + 1):
            if i != max_page:
                self._memes_on_page(i, self._posts_per_page)
            else:
                self._memes_on_page(i, num % self._posts_per_page)

    def _get_memes_helper(self, page_num):
        """ Helper function for the get_memes() function

        Return a list of memes on the specified page given. This function
        uses the API of the memegenerator.net
        """
        url = self._api + self._method_entry
        payload = {"languageCode": "en",
                   "pageIndex": page_num, "days": self._popular_days}
        r = requests.get(url, params=payload, timeout=self._timeout)

        try:
            json_memes = r.json()
        except ValueError as err:  # cannot decode json
            sys.stderr.write("ERROR: {} \n".format(str(err)))

        cmemes = json_memes["result"]
        meme_list = []

        for x in cmemes:

            # Picture url
            instance_image_url = ""
            try:
                instance_image_url = x["instanceImageUrl"]
            except KeyError:
                pass

            # Caption
            ccaption = x["text0"]
            try:
                extra = " --- " + x["text1"]
                ccaption += extra
            except TypeError:  # Returned json sometimes doesn't have text1
                pass
            except KeyError:  # Does not have the text1 tags
                pass

            # Raw image (meme macro) url
            imageUrl = ""
            try:
                imageUrl = x["imageUrl"]
            except KeyError:
                pass

            # Tags
            ctags = []
            try:
                ctags += [x["displayName"]]
            except KeyError:
                pass

            cscore = -1
            try:
                cscore = x["totalVotesScore"]
            except KeyError:
                pass

            cmeme = Meme(instance_image_url,
                         datetime.datetime.now(),
                         caption=ccaption,
                         raw_pic_url=imageUrl,
                         origin=self._origin,
                         tags=ctags,
                         score=cscore)
            meme_list.append(cmeme)

        return meme_list

    def _memes_on_page(self, page_num, n):
        """ Get num memes on page

        memegenerator.net has a convenient api that allows us to get memes
        in JSON format.
        API Documentation: http://version1.api.memegenerator.net/
        """

        if n > self._posts_per_page:
            return None

        # Use the helper function to get a list of memes on the page
        meme_list = self._get_memes_helper(page_num)

        for i in range(n):
            self._meme_pool.add(meme_list[i])
            self._meme_deque.appendleft(meme_list[i])

    def _filename(self):
        """ Override superclass _filename method

        The reason why we need to override is because for the
        memegenerator.net website, it ranks memes in different
        days duration: most popular in 1 day, 1 week and 1 month

        We want to have different caches for each case.
        """
        hashID = hashlib.sha1()
        hashID.update(repr(self._url).encode('utf-8'))
        hashID.update(repr(self._popular_days).encode('utf-8'))
        # Create a file name with hexdecimal representation of the SHA1
        # hash
        file_name = "cache_{:s}.memecache".format(hashID.hexdigest())
        return file_name

    def _write_data_tuple(self):
        """ Write all internal states to a data tuple
        """
        data = (self._url, self._max_tries, self._meme_pool,
                self._meme_deque, self._last_update, self._cache_size,
                self._maxcache_day, self._popular_days)
        return data

    def _read_data_tuple(self, t_data):
        """ Read in a data tuple and replace all the instance variables
        """
        self._url = t_data[0]
        self._max_tries = t_data[1]
        self._meme_pool = t_data[2]
        self._meme_deque = t_data[3]
        self._last_update = t_data[4]
        self._cache_size = t_data[5]
        self._maxcache_day = t_data[6]
        self._popular_days = t_data[7]

    def _read_update_time_from_cache(self):
        """ Read cache update time from cache file
        """
        dtuple = self._read_cache()

        return dtuple[4]


class RedditMemes(MemeSite):

    def __init__(self, cache_size=500, maxcache_day=1,
                 popular_type="Daily", timeout=20):
        """ The __init__ method for MemeGenerator class

        Args:
            cache_size (int): Number of memes stored as cache
            maxcache_day (int): Number of days until the cache expires
        """
        super(RedditMemes, self).__init__(
            "https://www.reddit.com/r/memes/", cache_size, maxcache_day)
        self._origin = Origins.REDDITMEMES

        # Client ID and user agent requested by Reddit API
        config = configparser.ConfigParser()

        cdir = os.path.dirname(os.path.realpath(__file__))
        config.read(os.path.join(cdir, 'config.ini'))

        self._client_id = config['Reddit']['ClientID']
        self._client_secret = config['Reddit']['ClientSecret']
        if self._client_secret == '':
            self._client_secret = None
        self._user_agent = config['Reddit']['UserAgent'].format(sys.platform)

        # Generate a Reddit instance
        self._reddit = praw.Reddit(client_id=self._client_id,
                                   client_secret=self._client_secret,
                                   user_agent=self. _user_agent)

        if self._no_cache() or self._cache_expired():
            self._build_cache()

    def get_memes(self, num):
        """ Get memes from Reddit /r/meme subreddit
        """
        if self._no_cache() or self._cache_expired():
            self._build_cache()
        else:
            self._update_with_cache()

        if self._cache_size >= num:
            return self._pop_memes(num)
        else:
            # Haven't found a way to get memes
            # in a specific range using PRAW
            results = self._reddit.subreddit('memes').hot(limit=num)
            meme_results = []

            for submission in results:

                # Get required properties for the memes
                ctitle = submission.title
                curl = submission.url

                cmeme = Meme(curl, datetime.datetime.now(),
                             title=ctitle,
                             origin=Origins.REDDITMEMES)
                meme_results.append(cmeme)

            return meme_results

    def _populate(self, num):
        """ Populate the meme pool and deque

        This method uses the reddit API wrapper PRAW library.
        """
        # Get num submissions
        results = self._reddit.subreddit('memes').hot(limit=num)
        # Save each submissions into the deque and pool
        for submission in results:

            # Get required properties for the memes
            ctitle = submission.title
            curl = submission.url

            cmeme = Meme(curl, datetime.datetime.now(),
                         title=ctitle,
                         origin=Origins.REDDITMEMES)
            self._meme_pool.add(cmeme)
            self._meme_deque.appendleft(cmeme)

    def _filename(self):
        """ Generate a unique filename for the RedditMemes cache file
        """
        hashID = hashlib.sha1()
        hashID.update(repr(self._url).encode('utf-8'))

        # Create a file name with hexdecimal representation of the SHA1 hash
        file_name = "cache_{:s}.memecache".format(hashID.hexdigest())
        return file_name

    def _write_data_tuple(self):
        """ Write all internal states to a data tuple
        """
        data = (self._url, self._max_tries, self._meme_pool,
                self._meme_deque, self._last_update, self._cache_size,
                self._maxcache_day)
        return data

    def _read_data_tuple(self, t_data):
        """ Read in a data tuple and replace all the instance variables
        """
        self._url = t_data[0]
        self._max_tries = t_data[1]
        self._meme_pool = t_data[2]
        self._meme_deque = t_data[3]
        self._last_update = t_data[4]
        self._cache_size = t_data[5]
        self._maxcache_day = t_data[6]

    def _read_update_time_from_cache(self):
        """ Read update time from cache
        """
        dtuple = self._read_cache()
        return dtuple[4]



from .memesites import RedditMemes

A = RedditMemes()

B = A.get_memes(3)

B[0].ocr_caption(method="Tesseract", thres=False,cfg="urban")
B[1].ocr_caption(method="Tesseract", thres=False, cfg="urban")
B[2].ocr_caption(method="Tesseract", thres=False, cfg="urban")

#B[2].ocr_caption(method="FontMatching")

print("Result:")
print(B[0].get_title())
print(B[0].get_caption())
print(B[1].get_title())
print(B[1].get_caption())
print(B[2].get_title())
print(B[2].get_caption())


# Testing beautiful soup
import bs4
import requests
import memesites
import hashlib

page = requests.get('http://www.quickmeme.com')
soup = bs4.BeautifulSoup(page.text, 'html.parser')

# Find a list of posts on the quickmeme page
list_of_posts = soup.find_all(class_="post-image")

# Access the text of each post
list_of_texts = [str(x['alt']) for x in list_of_posts]

# get urls of each post
list_of_urls = [str(x['src']) for x in list_of_posts]

# Test Meme hash and eq
A = memesites.Meme('1', '2')
B = memesites.Meme('1', '2')
C = memesites.Meme('1', '3')
print(len(set([A, B, C])))

# Test Meme md5 hashing
memeA = memesites.Meme('1', '2')
hashID = hashlib.md5()
print(repr(memeA))
hashID.update(repr(memeA).encode('utf-8'))
print(hashID.hexdigest())

#A = memesites.QuickMeme()
#print("Should be false:", A._cache_expired())
#A._memes_on_page(1, 5)

#A_tuple = A._write_data_tuple()
# A._save_cache()
#B = memesites.QuickMeme()
# B._read_cache()
#print(A_tuple == B._write_data_tuple())

# Test get memes
#A = memesites.QuickMeme()
#result = A.get_memes(3)
#result2 = A.get_memes(500)
#result = A.get_memes(20)

# Test memegenerator.net get memes
A = memesites.MemeGenerator()
result = A.get_memes(32)
print(result)



""" OCR Comparison Module
"""

from __future__ import print_function
from __future__ import division
from builtins import range
from past.utils import old_div
import os

wl_path = os.path.join(os.path.dirname(__file__),"dict/linuxwords.txt")
wl = open(wl_path, "r").read().upper().split("\n")

def ocr(path):
    """ DIY OCR from scratch
    """
    from .memeocr import rawocr
    import meme_get.ocr.parse as parse
    parse.bds, parse.ccr = rawocr(path)
    return parse.guesscaption()


def ocrTesseract(path, thres=True, cfg="urban"):
    """ Tesseract OCR
    """
    from .memeocr import tesseract_ocr
    return tesseract_ocr(path, thres=thres, cfg=cfg)


def evalresult(t):
    """ Evaluate the quality of an ocr result
    """
    puncs = ".,!?"
    t = t.replace("\n", " ")
    for p in puncs:
        t = t.replace(p, " " + p)
    t = t.split(" ")
    score = 0.0
    for i in range(0, len(t)):
        if t[i] in wl:
            score += 1.0
    print(old_div(score, len(t)))
    return old_div(score, len(t))

def ocrcomp(path, *args):
    """ Sort a list of ocr functions by their quality
    """
    results = []
    for f in args:
        results.append((f, f(path)))
    return sorted(results, key=lambda x: evalresult(x[1]), reverse=True)


from __future__ import print_function
from __future__ import division
# pylint: disable=C0103
# pair letters in words with similar spelling


from builtins import range
from past.utils import old_div
def wordpair(w1, w2):
    result = []
    l = min(len(w1), len(w2))
    while len(w1) > 0 and len(w2) > 0:
        # print result
        if w1[0] == w2[0]:
            result.append((w1[0], w2[0]))
            w1 = w1[1:]
            w2 = w2[1:]
        else:
            ws = [w1, w2]
            xs = [len(w1), len(w2)]
            for n in range(1, 5):
                for i1, i2 in [(0, 1), (1, 0)]:
                    found = False
                    for j in range(len(ws[i1])):
                        for k in range(j, min(j + n, len(ws[i2]))):
                            if ws[i1][j] == ws[i2][k]:

                                if abs(j - k) + (j + k) < abs(xs[i1] - xs[i2]) + (xs[i1] + xs[i2]):
                                    xs[i1] = j
                                    xs[i2] = k
                                found = True
                                break
                        if found == True:
                            break
            x1, x2 = xs

            result.append((w1[:x1], w2[:x2]))
            w1 = w1[x1:]
            w2 = w2[x2:]
    if len(w1) > 0 or len(w2) > 0:
        result.append((w1[:], w2[:]))
    return result

# pretty print result of wordpair


def printwordpair(wp):
    l1 = ""
    l2 = ""
    for w in wp:
        l1 += w[0] + "\t"
        l2 += w[1] + "\t"
    return l1 + "\n" + l2 + "\n"

# test wordpair with function f


def testwordpair(f):
    print(f(wordpair("president", "precedent")))
    print(f(wordpair("affection", "affectation")))
    print(f(wordpair("eminent", "immanent")))
    print(f(wordpair("principal", "principle")))
    print(f(wordpair("desert", "dessert")))
    print(f(wordpair("deed", "indeed")))
    print(f(wordpair("immense", "intense")))
    print(f(wordpair("drastic", "dramatic")))
    print(f(wordpair("emulsion", "emotion")))
    print(f(wordpair("wikipedia", "vvil<ieolix")))

# similarity of a word pair


def wordsim(wp):
    return sum([w[0] == w[1] for w in wp]) * 1.0 / sum([old_div((len(w[0]) + len(w[1])), 2.0) for w in wp])


if __name__ == "__main__":
    testwordpair(printwordpair)
    testwordpair(wordsim)



import praw

reddit = praw.Reddit(client_id='Ho_FIRMXbFiY3w', client_secret=None,
                     user_agent='linux:meme_get:v1.0 (by /u/memegenmemegen)')

results = reddit.subreddit('memes').hot(limit=10)

for a in results:
    print(a.title)
    print(a.url)



from __future__ import print_function
#pylint: disable=C0103

from builtins import range
from PIL import Image, ImageDraw, ImageFont
import memeocr as mo
import json


def charsim():
    result = {}
    cimgs = mo.cimgs
    C = mo.C

    for i in range(0, len(cimgs)):
        print(C[i])
        scores = []
        px1 = cimgs[i].load()
        for j in range(0, len(cimgs)):
            if i != j:
                score = 0
                px2 = cimgs[j].load()
                for x in range(0, cimgs[i].size[0]):
                    for y in range(0, cimgs[i].size[1]):
                        if x >= cimgs[j].size[0] or y >= cimgs[j].size[1]:
                            score -= 1
                        if px1[x, y][:3] == px2[x, y][:3]:
                            score += 1
                        else:
                            score -= 1
                scores.append((C[j], score))
        ns = mo.normalize(scores)
        result[C[i]] = {}
        for n in ns:
            result[C[i]][n[0]] = n[1]

    return result

if __name__ == "__main__":
    mo.makeglyphs()
    js = json.dumps(charsim(), sort_keys=True)
    f1 = open("similarity.json", "w")
    f1.write(js)



from __future__ import print_function
from __future__ import division
from builtins import range
from past.utils import old_div
import json
import itertools
import meme_get.ocr.util as util
import warnings
import os


# loads raw ocr data
#path = "images/img5.jpg"
# fi = open("data/" + path.split("/")[-1].split(".")[0] + ".json", "r")
#js = json.loads(fi.read())
bds = None
ccr = None
#bds = js[0]
#ccr = js[1]

# dictionary
#wl = open("dict/linuxwords.txt", "r").read().upper().split("\n")

wl_path = os.path.join(os.path.dirname(__file__),"dict/linuxwords.txt")
wl = open(wl_path, "r").read().upper().split("\n")


# format raw ocr data
def guessformat():
    result = ""
    lines = []
    for i in range(0, len(bds)):
        found = 0
        for j in range(0, len(lines)):
            if len(lines[j]) > 0:
                if abs(bds[i][1] - old_div(sum([bds[l][1] for l in lines[j]]), len(lines[j]))) < 5:
                    lines[j].append(i)
                    found = 1
                    break
        if found == 0:
            lines.append([i])

    for i in range(0, len(lines)):
        lines[i] = sorted(lines[i], key=lambda x: bds[x][0])
    lines = sorted(lines, key=lambda x: bds[x[0]][1])

    for j in range(0, len(lines)):
        l = lines[j]
        nl = []
        for i in range(0, len(l)):
            nl.append(l[i])
            if i < len(l) - 1:
                if bds[l[i + 1]][0] - bds[l[i]][2] > 0.3 * sum([bds[k][2] - bds[k][0] for k in l]) / len(l):
                    nl.append(" ")
            if bds[l[i]][3] - bds[l[i]][1] < 0.5 * sum([bds[k][3] - bds[k][1] for k in l]) / len(l):
                nl.pop()
                nl.append("'")
            elif bds[l[i]][3] - bds[l[i]][1] < 0.9 * sum([bds[k][3] - bds[k][1] for k in l]) / len(l):
                nl.pop()
                if ccr[l[i]][0][0] == "I" or ccr[l[i]][0][0] == "J":
                    nl.append("!")
                else:
                    nl.append("?")
        lines[j] = nl

    return lines


# guess a line of text
def guessline(line, simple=False):
    words = [[]]
    for i in line:
        if i == " ":
            words.append([])
        elif i == "?" or i == "!":
            words.append(i)
        else:
            if type(words[-1]) == list:
                words[-1].append(i)
            else:
                words.append([i])
    for i in range(0, len(words)):
        if type(words[i]) == list:
            words[i] = guessword(words[i], simple)
    return " ".join(words)

# sort possible corrections


def sortchange(word):
    changes = []
    for w in word:
        if type(w) == int:
            for i in range(1, len(ccr[w])):
                changes.append((w, ccr[w][i]))

    changes = sorted(changes, key=lambda x: x[1][1])
    return changes


# guess a word
def guessword(w, simple=False):
    word = w[:]
    changes = {}
    pot = sortchange(word)

    guess = ""
    for i in range(0, len(word)):
        if type(word[i]) == int:
            word[i] = (word[i], ccr[word[i]][0][0])
        else:
            word[i] = (-1, word[i])

    guess = "".join([w[1] for w in word])
    guess0 = guess

    if simple:
        return guess0

    def ind(word, i):
        for j in range(0, len(word)):
            if word[j][0] == i:
                return j
        return -1

    while guess not in wl:
        if len(pot) == 0 or pot[-1][1][1] < 0.8:
            return guess0
        nc = pot.pop()
        if nc[0] not in list(changes.keys()):
            changes[nc[0]] = []
        changes[nc[0]].append(nc)

        chco = util.allchoice(list(changes.values()))

        for c in chco:
            newword = word[:]
            for d in c:

                newword[ind(newword, d[0])] = (d[0], d[1][0])

            guess = "".join([w[1] for w in newword])
            # print guess
            if guess in wl:
                return guess

    return guess

# check if some text is noise


def isgibber(t):
    if len(t) < 3:
        return True
    if len(t) < 4 and " " in t:
        return True
    if t.count(" ") > old_div(len(t), 3):
        return True

    return False

# guess the caption of a meme


def guesscaption(simple=False):
    output = ""
    gf = guessformat()

    print("raw: ")
    for g in gf:
        print(guessline(g, True))
    print()

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        for g in gf:
            gl = guessline(g, simple=simple)
            if not isgibber(gl):
                output += gl.replace(" !", "!").replace(" ?", "?") + "\n"

    return output

if __name__ == "__main__":
    print(guesscaption())



"""meme_get setup.py"""
from os import path
from setuptools import setup, find_packages

PACKAGE_NAME = 'meme_get'
HERE = path.abspath(path.dirname(__file__))

with open(path.join(HERE, 'README.rst'), encoding='utf-8') as f:
    README = f.read()


setup(name=PACKAGE_NAME,
      version='1.0.0b1',
      description='Get memes',
      long_description=README,
      license='MIT',
      author='Jingnan Shi',
      author_email='jshi@g.hmc.edu',
      url='https://github.com/memegen/meme_get',
      classifiers=[
          'Development Status :: 4 - Beta',
          'Environment :: Console',
          'Intended Audience :: Developers',
          'Natural Language :: English',
          'Operating System :: OS Independent',
          'License :: OSI Approved :: MIT License',
          'Programming Language :: Python',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 3.5',
          'Topic :: Utilities'
      ],
      keywords=['meme', 'client', 'download'],
      install_requires=[
          'beautifulsoup4',
          'requests',
          'pyenchant',
          'praw',
          'lxml',
          'Pillow'],
      packages=find_packages(exclude=['tests', 'tests.*']),
      package_data={
          '':['LICENSE', '*.rst', '*.txt', '*.ini', '*.data', '*.ttf', '*.cfg']
      })



from __future__ import print_function
from __future__ import division

from builtins import range
from past.utils import old_div
import random
import json
import sys
import os
import enchant
import pyocr
import pyocr.builders
from PIL import Image, ImageDraw, ImageFont

path = "images/img8.jpg"

# characters to recognize
C = "ABCDEFGHIJKLMNOPQRSTUVWXYZ123456789"

# standard character glyphs
cimgs = []

# original image
IM = None
w, h = 0, 0
PX = None

# display layer
disp = None
draw = None

# threshold layer
thim = None
thpx = None
thdr = None

# character areas
areas = []
badareas = []


def loadimg(path):
    global IM, w, h, PX, disp, draw, thim, thpx, thdr
    # original image
    IM = Image.open(path)
    w, h = IM.size
    PX = IM.load()

    # display layer
    disp = Image.new("RGB", (w, h))
    draw = ImageDraw.Draw(disp)

    # threshold layer
    thim = Image.new("RGB", (w, h))
    thpx = thim.load()
    thdr = ImageDraw.Draw(thim)

def closeimg():
    """ Close all the image
    """
    IM.close()
    disp.close()
    thim.close()

# rgb(255,255,255) to hsv(360,1.0,1.0) conversion
def rgb2hsv(r, g, b):
    R = old_div(r, 255.0)
    G = old_div(g, 255.0)
    B = old_div(b, 255.0)
    cmax = max(R, G, B)
    cmin = min(R, G, B)
    delta = cmax - cmin
    H, S, V = 0, 0, 0
    if delta == 0:
        H = 0
    elif cmax == R:
        H = 60 * ((old_div((G - B), delta)) % 6)
    elif cmax == G:
        H = 60 * (old_div((B - R), delta) + 2)
    elif cmax == B:
        H = 60 * (old_div((R - G), delta) + 4)
    if cmax == 0:
        S = 0
    else:
        S = old_div(delta, cmax)
    V = cmax
    return H, S, V

# find the left cutoff of character glyph


def firstwhitex(im):
    fwx = im.size[0]
    px = im.load()
    for y in range(0, im.size[1]):
        for x in range(0, im.size[0]):
            r, g, b = px[x, y][:3]
            if (r, g, b) == (255, 255, 255):
                if x < fwx:
                    fwx = x
    return fwx

# been there


def visited(x, y, areas):
    for i in range(0, len(areas)):
        if (x, y) in areas[i]:
            return True
    return False

# floodfill from pixel


def flood(x, y, d):
    global areas
    area = []
    seeds = [(x, y)]
    while d > 0 and len(seeds) > 0:

        d -= 1
        if d == 0:
            # print "too large"
            badareas.append(area)
            return []
        x, y = seeds.pop()

        if visited(x, y, areas) or visited(x, y, badareas):
            # print "visited"
            return []

        if (d > 0 and x > 0 and x < w - 1 and y > 0 and y < h - 1):
            if not((x, y) in area):
                r, g, b = thpx[x, y][:3]

                if (r != 0):
                    area.append((x, y))
                    seeds += [(x, y), (x - 1, y), (x, y - 1),
                              (x + 1, y), (x, y + 1)]
    return area


# get all character areas
def getareas():
    for y in list(range(0, int(old_div(h, 4)), 10)) + list(range(int(3 * h / 4), h, 10)):
        print(y, "/", h)
        for x in range(0, w, 5):
            area = flood(x, y, 30000)
            # print area
            if len(area) > 1:
                col = (random.randrange(0, 255), random.randrange(
                    0, 255), random.randrange(0, 255))
                for i in range(0, len(area)):
                    draw.point(list(area[i]), fill=col)
                areas.append(area)

# boundaries of a character


def getbounds(area):
    xmin = area[0][0]
    xmax = area[0][0]
    ymin = area[0][1]
    ymax = area[0][1]
    for i in range(0, len(area)):
        if area[i][0] < xmin:
            xmin = area[i][0]
        if area[i][1] < ymin:
            ymin = area[i][1]
        if area[i][0] > xmax:
            xmax = area[i][0]
        if area[i][1] > ymax:
            ymax = area[i][1]
    return xmin - 1, ymin - 1, xmax + 1, ymax + 1

# draw a boundary


def drawbounds():
    bds = []
    for i in range(0, len(areas)):
        bd = getbounds(areas[i])
        bds.append(bd)
        draw.rectangle(bd, outline=(255, 0, 0))
    return bds


# OCR
def checkchars():
    scoreboard = []
    for i in range(0, len(areas)):
        print(i, "/", len(areas))
        bd = getbounds(areas[i])
        scores = []
        
        print(len(cimgs))

        for j in range(0, len(cimgs)):
            score = 0
            px = cimgs[j].load()
            score = 0
            sc = old_div((1.0 * (bd[3] - bd[1])), 100.0)
            for x in range(0, cimgs[j].size[0]):
                for y in range(0, cimgs[j].size[1]):
                    xp = min(int(x * sc + bd[0]), w - 1)
                    yp = min(int(y * sc + bd[1]), h - 1)

                    r1, g1, b1 = px[x, y][:3]
                    r2, g2, b2 = thpx[xp, yp][:3]

                    if (xp < bd[2]):
                        if (r1 == r2):
                            score += 1
                        else:
                            score -= 1
                    else:
                        if (r1 == 0):
                            score += 0
                        else:
                            score -= 1
            scores.append((C[j], old_div(int(score * 10), 10.0)))
        scoreboard.append(normalize(scores))
        draw.text((bd[0], bd[1] - 5), normalize(scores)[0][0], (0, 255, 255))
    return scoreboard


# normalize scores to 0.0-1.0
def normalize(scores):
    scores = sorted(scores, key=lambda x: x[1], reverse=True)

    ns = sorted(scores, key=lambda x: x[1], reverse=True)

    for i in range(0, len(scores)):

        if scores[i][1] <= 0 or scores[0][1] == 0:
            ns[i] = (scores[i][0], 0)
        else:
            n = old_div((scores[i][1] * 1.0), scores[0][1])
            ns[i] = (scores[i][0], old_div(int(n * 1000), 1000.0))

    return ns

# print OCR result


def showresult(scoreboard):
    for i in range(0, len(scoreboard)):
        print("".join([s[0] for s in scoreboard[i]]))

# make character glyphs


def makeglyphs():

    for i in range(0, len(C)):
        im = Image.new("RGB", (100, 110))
        dr = ImageDraw.Draw(im)
        font = ImageFont.truetype(os.path.join(os.path.dirname(__file__),"fonts/Impact.ttf"), 124)
        dr.text((0, -25), C[i], (255, 255, 255), font=font)

        fwx = firstwhitex(im)

        im = Image.new("RGB", (100, 110))
        dr = ImageDraw.Draw(im)
        font = ImageFont.truetype(os.path.join(os.path.dirname(__file__),"fonts/Impact.ttf"), 124)
        dr.text((-fwx, -26), C[i], (255, 255, 255), font=font)

        cimgs.append(im)

# make threshold image


def thresh():
    for x in range(0, w):
        for y in range(0, h):
            r, g, b = PX[x, y][:3]
            hsv = rgb2hsv(r, g, b)
            if (hsv[2] > 0.9)and hsv[1] < 0.1:
                thdr.point([x, y], fill=(255, 255, 255))
            else:
                thdr.point([x, y], fill=(0, 0, 0))

# returns possible characters and bounds in an image


def rawocr(path):
    global cimgs
    print("Starting ocr for {}".format(str(path)))
    # Clear glyphs
    cimgs = []

    loadimg(path)
    print("p0: ", len(cimgs))
    makeglyphs()

    print("p1: ", len(cimgs))
    thresh()
    # thim.show()
    getareas()
    bds = drawbounds()

    # disp.show()
    print("p2: ", len(cimgs))

    ccr = checkchars()
    showresult(ccr)
    # disp.show()
    closeimg()
    print("Finish OCR.")
    return bds, ccr


def tesseract_ocr_helper(base_image, config="Default"):
    """ A wrapper for using tesseract to do OCR
    """
    tools = pyocr.get_available_tools()
    if len(tools) == 0:
        print("No OCR tool found")
        sys.exit(1)

    # The tools are returned in the recommended order of usage
    tool = tools[0]
    print("Will use tool '%s'" % (tool.get_name()))

    langs = tool.get_available_languages()
    print("Available languages: %s" % ", ".join(langs))
    lang = langs[0]
    print("Will use lang '%s'" % (lang))

    custom_builder = pyocr.builders.TextBuilder()
    if config != "Default":
        custom_builder.tesseract_configs = [config]

    txt = tool.image_to_string(
        base_image,
        lang=lang,
        builder=custom_builder
    )

    # Spell correct
    dict_path = os.path.join(os.path.dirname(__file__),"dict/urban_dict.txt")
    d = enchant.DictWithPWL("en_US", dict_path)
    txtA = txt.replace('\n', ' \n ')
    A = txtA.split(" ")
    B = []

    for x in A:
        if (x != '\n' and len(x) != 0
                and d.check(x) is False
                and len(d.suggest(x)) != 0):
            B.append(d.suggest(x)[0])
        else:
            B.append(x)

    return " ".join(B)


def tesseract_ocr(path, thres=False, cfg="Default"):
    """ Wrapper for tesseract OCR
    """
    loadimg(path)
    if thres:
        thresh()
        result = tesseract_ocr_helper(thim, config=cfg)
        return result
    else:
        result = tesseract_ocr_helper(IM, config=cfg)
        return result


if __name__ == "__main__":
    bds, ccr = rawocr(path)

    js = json.dumps([bds, ccr])
    fo = open("data/" + path.split("/")[-1].split(".")[0] + ".json", "w")
    fo.write(js)



from PIL import Image, ImageDraw


def rgb2hsv(r, g, b):
    R = r / 255.0
    G = g / 255.0
    B = b / 255.0
    cmax = max(R, G, B)
    cmin = min(R, G, B)
    delta = cmax - cmin
    H, S, V = 0, 0, 0
    if delta == 0:
        H = 0
    elif cmax == R:
        H = 60 * (((G - B) / delta) % 6)
    elif cmax == G:
        H = 60 * ((B - R) / delta + 2)
    elif cmax == B:
        H = 60 * ((R - G) / delta + 4)
    if cmax == 0:
        S = 0
    else:
        S = delta / cmax
    V = cmax
    return H, S, V


def thres(pil_Image):
    """ Thresholding the image
    """
    w, h = pil_Image.size
    thim = Image.new("RGB", (w, h))
    thdr = ImageDraw.Draw(thim)
    PX = pil_Image.load()
    # make threshold image
    for x in range(0, w):
        for y in range(0, h):
            r, g, b = PX[x, y][:3]
            hsv = rgb2hsv(r, g, b)
            if hsv[2] > 0.9 and hsv[1] < 0.1:
                thdr.point([x, y], fill=(0, 0, 0))
            else:
                thdr.point([x, y], fill=(255, 255, 255))
    thim.show()
    return thim
